// Diseño de la calculadora con un sólo bus

//
// Unidad de datos
//
// Se diseñan todos los componentes de la ruta de datos en un
// mismo módulo.

module calculator_dp #(
    parameter W = 8     // anchura del bus
    )(
    input wire clk,     // señal de reloj
    input wire ra,      // lectura de registro A
    input wire wa,      // escritura en registro A
    input wire rb,      // lectura de registro B
    input wire wb,      // escritura en registro B
    input wire p1, p0,  // selección de operación ALU
    input wire wt,      // escritura en registro tampón
    input wire rac,     // lectura del acumulador
    input wire wac      // escritura en el acumulador
    );

    wire [W-1:0] bus;   // bus
    reg [W-1:0] out;    // salida ALU
    reg [W-1:0] a, b, t, ac;    // registros

    // registro A
    always @(posedge clk)
        if (wa)
            a <= bus;

    assign bus = ra ? a : 'bz;

    // registro B
    always @(posedge clk)
        if (wb)
            b <= bus;

    assign bus = rb ? b : 'bz;

    // registro T
    always @(posedge clk)
        if (wt)
            t <= bus;

    // registro AC
    always @(posedge clk)
        if (wac)
            ac <= out;

    assign bus = rac ? ac : 'bz;

    // ALU
    always @(*)
        case({p1, p0})
        2'b00:
            out = bus + t;
        2'b01:
            out = bus;
        2'b10:
            out = bus - t;
        default:
            out = t;
        endcase

endmodule // calculator_dp

//
// Unidad de control
//
// Se diseña con tres procedimientos:
//
//   * Cambio de estado (secuencial)
//   * Próximo estado (combinacional)
//   * Cálculo de la salida (combinacional)

module calculator_ctl (
    // señales externas
    input wire clk,     // señal de reloj
    input wire reset,   // señal de reset
    input wire xs,	    // señal de inicio
    input wire i1, i0,  // señales de operación
    output reg finish,  // señal fin de operación
    // señales hacia la unidad de datos
    output reg ra,      // lectura de registro A
    output reg wa,      // escritura en registro A
    output reg rb,      // lectura de registro B
    output reg wb,      // escritura en registro B
    output reg p1, p0,  // selección de operación ALU
    output reg wt,      // escritura en registro tampón
    output reg rac,     // lectura del acumulador
    output reg wac      // escritura en el acumulador
    );

    // estado y próximo estado
    reg [2:0] state, next_state;

    // codificación de estados
    parameter s0 = 0,
              s1 = 1,
              s2 = 2,
              s3 = 3,
              sf = 7;

    // procedimiento de cambio de estado
    always @(posedge clk, posedge reset)
        if (reset)
            state <= s0;
        else
            state <= next_state;

    // cálculo del próximo estado
    always @(*)
        case(state)
        s0: next_state = xs ? s1 : s0;
        s1: next_state = s2;
        s2: next_state = s3;
        s3: next_state = sf;
        sf: next_state = s0;
        endcase

    // cálculo de la salida
    always @(*) begin
        // valor por defecto 0 para todas las salidas
        finish = 1'b0;
        ra = 1'b0;
        wa = 1'b0;
        rb = 1'b0;
        wb = 1'b0;
        p1 = 1'b0;
        p0 = 1'b0;
        wt = 1'b0;
        rac = 1'b0;
        wac = 1'b0;

        case(state)
        s1: begin
            // T <- B
            rb = 1'b1; wt = 1'b1;
        end
        s2: begin
            // (i1==0): AC<-A+T; (i1==1): AC<-A-T
            ra = 1'b1; wac = 1'b1; p1 = i1;
        end
        s3: begin
            // (i0==0): A<-AC; (i0==1): B<-AC
            rac = 1'b1;
            if(i0)
                wb = 1'b1;
            else
                wa = 1'b1;
        end
        sf:
            finish = 1'b1;
        endcase
    end

endmodule // calculator_ctl

//
// Caculadora completa
//
// Simplemente se conectan la unidad de datos y de control

module calculator #(
    parameter W = 8
    )(
    input wire clk,     // señal de reloj
    input wire reset,   // señal de reset
    input wire xs,	    // señal de inicio
    input wire i1, i0,  // señales de operación
    output wire finish  // señal fin de operación
    );

    // señales internas
    wire ra, wa, rb, wb, p1, p0, wt, rac, wac;

    calculator_ctl control(
        .clk(clk),
        .reset(reset),
        .xs(xs),
        .i1(i1),
        .i0(i0),
        .finish(finish),
        .ra(ra),
        .wa(wa),
        .rb(rb),
        .wb(wb),
        .p1(p1),
        .p0(p0),
        .wt(wt),
        .rac(rac),
        .wac(wac)
        );

    calculator_dp #(.W(W)) data_path (
        .clk(clk),
        .ra(ra),
        .wa(wa),
        .rb(rb),
        .wb(wb),
        .p1(p1),
        .p0(p0),
        .wt(wt),
        .rac(rac),
        .wac(wac)
        );

endmodule // calculator
